// ======================================================================
//
// camera.h
// jeff grills
//
// copyright 1996 origin systems inc
//                                                                        
// ======================================================================

#ifndef CAMERA_H
#define CAMERA_H

// ======================================================================
// includes

class    Light;
class    Material;
class    ObjectList;

#include "Object.h"
#include "Gl.def"

// ======================================================================

class Camera : public Object
{
public:

	struct RenderStats
	{
		int      objectsRendered;
		int      objectsCulled;
		int      lightsRendered;
		int      lightsCulled;
		int      passedSphereTests;
		int      failedSphereTests;

		int      setupCalls;
		int      setupGLCalls;
		int      projectCalls;
		int      matrixMultiplies;
		
		int      verticesTransformed;

		int      polygonTotalCount;
		int      polygonRenderCount;
		int      polygonRenderArea;
		int      polygonBackfacingCount;
		int      polygonClipInitialOutCount;
		int      polygonClipOutCount;
		int      polygonClipCount;
		int      polygonClipEdgeCount;
		int      polygonUnclippedCount;
		int      polygonSplitCount;

		int      polygonsPerspective16;
		int      polygonsPerspective8;
		int      polygonsPerspective4;
		int      polygonsAffine;

		int      translucentPolygonsDrawn;

		void reset(void);
		void debugDump(void);
	};

	static RenderStats *renderStats;

public:

	enum Space
	{
		WorldSpace,
		CameraSpace
	};

protected:

	enum
	{
		DEFAULT_OBJECT_LISTS            = 10,
		FRUSTUM_PLANES                  = 6
	};

private:

	struct RenderObject
	{
		const Appear  *appear;
		const Xform   *objectToWorld;
		RenderObject  *next;
	};

	static RenderStats  renderStatsObject;

	// list of lights in the scene
	static int           renderLightMax;
	static int           renderLightCount;
	static Light       **renderLight;

	// list of objects in the scene
	static bool          renderObjectLocked;
	static int           renderObjectCount;
	static int           renderObjectMax;
	static RenderObject *renderObject;

private:

	void renderExtents(RenderObject *list);
	void sortTransformClipAndProjectObjects(RenderObject *first);

protected:

	// viewport data
	int          x0, y0, x1, y1;
	int          halfWidth, halfHeight;
	int          centerX, centerY;

	// near and far plane data and field-of-view
	real         nearPlane;
	real         farPlane;
	real         fieldOfView;

	real         clipWidthMultiplier, clipHeightMultiplier;
	real         viewportWidthMultiplier, viewportHeightMultiplier;

	// camera space description of the viewing frustum
	Vector       frustumNormal[FRUSTUM_PLANES];
	real         frustumPlaneD[FRUSTUM_PLANES];

	// object lists to render
	int                  maxObjectLists;
	int                  numObjectLists;
	const ObjectList   **objectList;
	const Object        *excludeObject;


private:
	
	// disable the copy constructor and the assignment operators
	Camera(const Camera &);
	Camera &operator =(const Camera &);

private:

	void setup(void);
	
public:

	static void install(void);
	static void remove(void);

public:

	explicit Camera(int newMaxObjectLists=DEFAULT_OBJECT_LISTS);
	virtual ~Camera(void);

	void debugDump(void);

	// setup camera	
	void         setPerspectiveProjection(void);
	void         setNear(real nearDist);
	void         setFar(real farDist);
	virtual void setViewport(int vx0, int vy0, int vx1, int vy1);
	void         setFieldOfView(real fovRadians);

	// perspective projection of a point into the viewing frustum
	bool         project(const Vector &vector, real *x, real *y, Space space=WorldSpace, bool offscreenAllowed=false) const;
	bool         project(const Vector &vector, int *x, int *y, Space space=WorldSpace, bool offscreenAllowed=false) const;

	// render routines
	void         setExclude(const Object *newExcludeObject);
	bool         isExcluded(const Object *newExcludeObject);
	void         addObjectList(const ObjectList *objList);
	bool         testSphereVisibility(const Vector &worldPoint, real radius, int testPlanes=CLIP_ALL) const;
	void         removeObjectList(const ObjectList *objList);
	void         addObject(const Object *object, const Appear *appear);
	void         addLight(Light *light);
	void         setupGL(void);
	virtual bool render(bool needSetupGL);

	// routines for use by geometry engine
	void          getViewport(int *vx0, int *vy0, int *vx1, int *vy1) const;
	real          getNearPlane(void) const;
	real          getFarPlane(void) const;
	int           getCenterX(void) const;
	int           getCenterY(void) const;
	real          getClipWidthMultiplier(void) const;
	real          getClipHeightMultiplier(void) const;
	real          getWidthMultiplier(void) const;
	real          getHeightMultiplier(void) const;
	const Vector &getFrustumNormal(int i) const;
	real          getFrustumPlaneD(int i) const;
	real          getFieldOfView(void) const;
};

// ======================================================================

inline void Camera::setExclude(const Object *newExcludeObject)
{
	excludeObject = newExcludeObject;
}

// ----------------------------------------------------------------------

inline bool Camera::isExcluded(const Object *testExcludeObject)
{
	return (excludeObject == testExcludeObject);
}

// ----------------------------------------------------------------------

inline real Camera::getNearPlane(void) const
{
	return nearPlane;
}

// ----------------------------------------------------------------------

inline real Camera::getFarPlane(void) const
{
	return farPlane;
}

// ----------------------------------------------------------------------

inline int Camera::getCenterX(void) const
{
	return centerX;
}

// ----------------------------------------------------------------------

inline int Camera::getCenterY(void) const
{
	return centerY;
}

// ----------------------------------------------------------------------

inline real Camera::getClipWidthMultiplier(void) const
{
	return clipWidthMultiplier;
}

// ----------------------------------------------------------------------

inline real Camera::getClipHeightMultiplier(void) const
{
	return clipHeightMultiplier;
}

// ----------------------------------------------------------------------

inline real Camera::getWidthMultiplier(void) const
{
	return viewportWidthMultiplier;
}

// ----------------------------------------------------------------------

inline real Camera::getHeightMultiplier(void) const
{
	return viewportHeightMultiplier;
}

// ----------------------------------------------------------------------

inline const Vector &Camera::getFrustumNormal(int i) const
{
	return frustumNormal[i];
}

// ----------------------------------------------------------------------

inline real Camera::getFrustumPlaneD(int i) const
{
	return frustumPlaneD[i];
}

// ----------------------------------------------------------------------

inline real Camera::getFieldOfView(void) const
{
	return fieldOfView;
}

// ----------------------------------------------------------------------

inline void Camera::getViewport(int *vx0, int *vy0, int *vx1, int *vy1) const
{
	*vx0 = x0;
	*vy0 = y0;
	*vx1 = x1;
	*vy1 = y1;
}

// ======================================================================

#endif
